/*
 * strtoken.c : Enhanced strtok that can handle empty fields, returning
 *              a zero length string.
 *
 *              While it is meant to be a drop in replacement for strtok,
 *              the fact that it does not eat adjacent seperators may cause
 *              undesirable behaviour in some programs not expecting this.
 *
 *
 *              Empty Field Handling:
 *              
 *              ",xxx,,xx," results in "" "xxx" "" "xx" if "," is the sep
 *              "xxx,\n" results in "xxx" "" if ",\n" is the sep
 *
 *              This is deliberate to make reading delimitered files nice.
 *
 *
 *              Design Notes:
 *
 *              strtoken does *not* write on its own input, but makes
 *              a copy for the returned token (major criticism of strtok).
 *
 *
 *              Performance:
 *
 *              It is about 1.2 times slower than the  strtok.
 *
 *
 *				threading
 *
 *				This is not threadsafe (just like strtok), but it should be 
 *				simple to rework into a threadsage strtoken_r() or similar.
 *
 *
 * Copyright Mark Kirkwood (markir@paradise.net.nz)
 * This is open source software, and may be redistributed provided this 
 * notice is preserved. No warrenty is intended or implied.
 */
#include <stdio.h>
#include <stdlib.h>
#include "bool.h"


/*
 * This is faster than strlen().
 */
int		stringlen(const char *str) {

	int 	len;

	for (len = 0; *str++ != '\0'; len++) ; 

	return len;
} 


char	*strtoken(char *str, const char *sep) {

	static	char	*strbuf;		/* saved pointer to str */
	static	char	*strretbuf;		/* token to be returned */
	static	int		offset;			/* progress offset through the strbuf */
	static	int		strlength;		/* stash the length of str */
									/* minimize stringlen() calls */

	int				i, j, k;
	bool			match;			/* match sep flag */

	/*
	 * Initialize, save a pointer to the string and
	 * allocate a  buffer for the returning token.
	 */
	if ( str != NULL) {
		
		offset = 0;

		strlength = stringlen(str);

		strbuf = str;	
		strretbuf = malloc(strlength + 1);

		if (strretbuf == NULL ) {
			fprintf(stderr, "strtoken : error, out of memory\n");
			return NULL;
		}


	} 


	/*
	 * If we past in a NULL at initialization or
	 * read through the entire string, finish up.
	 */
	if ( strretbuf == NULL) {

		return NULL;

	} else if (offset == strlength) {

		free(strretbuf);

		return NULL;

	}

	
	/*
	 * For each element of the string, starting at offset, compare it to each
	 * of the seperator bytes in turn, stop when there is a match
	 *
	 * If a match was found then set the offset to record this, and
	 * null terminate the return buffer (and handle all the corner cases).
	 *
	 * Copy strbuf into strretfbuf byte by byte, privided they are not a sep.
	 */

	match = false;
	k = 0;

	for (i = offset; i < strlength ; i++) {
		
		/* Does the strbuf[i] element match any of the seps?*/
		j = 0;
		while ( sep[j] != '\0' ) {

			if (strbuf[i] == sep[j]) {			/* got a match, set the flag */
				match = true;
				break;
			}
			j++;
		}


		/* Did we see a match? */
		if (match) {							

			if ( k == 0) {

				/* Transform empty elements (adjacent seps) */
				strretbuf[0] = '\0';
				offset = offset + 1;
				break;

			} else {
				
				/* The normal case */
				strretbuf[k] = '\0';
				offset = i + 1;
				break;

			}
		}

		
		/* Continue to build up the token for return */
		strretbuf[k] = strbuf[i];
		k++;
		
	}


	/*
	 * No part of the string matched any of the sep's.
	 */
	if (!match) {

		offset = strlength;
		strretbuf[k] = '\0';

	}
	
	return strretbuf;

}
